/*
    This file is part of SNAKE.

    SNAKE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    SNAKE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

// DIR = E | S | W | N
// LEN = Lnge in Einheiten von Raster
// LEN % Raster muss 0 sein
// Bits
// 0       = Richtung
// 7,2..1  = Level-Filter
// 6..3    = Lnge in {1..15}

#define RASTER_LOG 3
#define RASTER (1 << RASTER_LOG)

/*
  Codierung

  LEN | DIR | LEV  |  Bedeutung
  ----+-----+------+----------------
   >0 |  *  | *    |  Mauer-Segment
    0 |  0  | *    |  END
    0 |  1  | >= 1 |  NEXT WALL
    0 |  1  | 0x40 |  GOTO 
*/

#define VAL(DIR,LEN,LEVEL) (0                           \
        | (/*0b00000001*/ 0x01 & (DIR >> 1))            \
        | (/*0b01111000*/ 0x78 & (LEN << RASTER_LOG))   \
        | (/*0b10000110*/ 0x86 & ((LEVEL) << 1))  )

#define SEG_1234(DIR,LEN)    VAL(SN_ ## DIR, LEN, 1-1) // >= 1
#define SEG__234(DIR,LEN)    VAL(SN_ ## DIR, LEN, 2-1) // >= 2
#define SEG___34(DIR,LEN)    VAL(SN_ ## DIR, LEN, 3-1) // >= 3
#define SEG____4(DIR,LEN)    VAL(SN_ ## DIR, LEN, 4-1) // >= 4

#define SEG_123_(DIR,LEN)    VAL(SN_ ## DIR, LEN, 3 | 0x40) // <= 3
#define SEG_12__(DIR,LEN)    VAL(SN_ ## DIR, LEN, 2 | 0x40) // <= 2
#define SEG_1___(DIR,LEN)    VAL(SN_ ## DIR, LEN, 1 | 0x40) // <= 1
#define SEG_____             VAL(         0,   1, 0 | 0x40) // <= 0 Nie

#define WALL_GOTO         VAL(2,0,0x40)
#define WALL_NEXT(LEVEL)  VAL(2,0,LEVEL-1)
#define WALL_END  0

// Das erste Element muss nach Osten
// Das erste Element muss fuer alle Level aktiv sein
const uint8_t WALL[] PROGMEM =
{
    WALL_GOTO, 44, 52, // = 4 % 8
        SEG_1234 (E,  6),
        SEG__234 (N,  3),
        SEG_1234 (E,  8),
        SEG____4 (N,  2),
        SEG____4 (E,  1),
        SEG__234 (S,  3),
        SEG____4 (W,  8),
        SEG____4 (S,  2),
        SEG____4 (E,  7),
        SEG_____        ,
        SEG_1234 (E,  6),
        
        SEG_1234 (N, 10),
        SEG__234 (W,  1),
        SEG_1234 (N,  2),
        SEG__234 (E,  1),
        SEG_1234 (N,  4),
        
        SEG_1234 (W, 15),
        SEG_____        ,
        SEG_1___ (W,  5),
        SEG__234 (S,  5),
        
        SEG__234 (W,  5),
        SEG_1___ (S,  5),
        SEG_____        ,
        SEG_1234 (S, 11),
    WALL_NEXT (3),
    
    WALL_GOTO, 148, 140,
        SEG___34 (E,  3),
        SEG___34 (S,  1),
        SEG___34 (W,  1),
        SEG___34 (S,  4),
        SEG___34 (W,  1),
        SEG___34 (N,  4),
        SEG___34 (W,  1),
        SEG___34 (N,  1),
    WALL_END
};

const struct 
{
    struct { uint8_t x, y; } start;
    uint8_t zoom;
} snake_text_init[] PROGMEM =
{
    [0] = { .start = {100, 220}, .zoom  = 1},
    [1] = { .start = {100, 25},  .zoom  = 1},
    [2] = { .start = {70,   -1}, .zoom  = 3}
};

static const snake_level_t SNAKE_LEVELS[] PROGMEM = 
{
    [1-1] = 
    {
        .delay  = 3,
        .grow   = 200,
        .raster = (1 << 3),
    }
    ,
    [2-1] = 
    {
        .delay  = 1,
        .grow   = 200,
        .raster = (1 << 3),
    }
    ,
    [3-1] = 
    {
        .delay  = 1,
        .grow   = 200,
        .raster = (1 << 3),
    }
    ,
    [4-1] = 
    {
        .delay  = 1,
        .grow   = 200,
        .raster = (1 << 3),
    }
};

// Zum Zeichnen von Kopf + Futter
static const struct { int8_t dx, dy;} snake_head_dxy[] PROGMEM = 
{
    {-1, -1},    {-1,  0},    {-1,  1},     {0,  1},     {1,  1},
    { 1,  0},    { 1, -1},    { 0, -1},     {0,  0}
};

// Eigene Glyphs fr VEKFONT

#define RIGHT1 | (1 << 5)

const uint8_t snake_glyph[] PROGMEM =
{
    // Kopf
    '$', 13
    , 4, 3
    , VF_LINE (2,  -1,  0, 0, 1)
    , VF_LINE (1,  -1,  1, 0, 1)
    , VF_LINE (1,   0,  1, 0, 1)
    , VF_LINE (1,   1,  1, 0, 1)
    , VF_LINE (2,   1,  0, 0, 1)
    , VF_LINE (1,   1, -1, 0, 1)
    , VF_LINE (1,   0, -1, 0, 1)
    , VF_LINE (1,  -1, -1, 0, 1)
    
    , VF_LINE (2,   0, -1, 0, 1)
    , VF_LINE (1,   1, -1, 0, 1)
    , VF_LINE (1,   1,  0, 0, 1)
    ,
    // Krper
    '~', 4  RIGHT1
    , VF_LINE (2,   1,  1, 1, 1)
    , VF_LINE (1,   1,  0, 0, 1)
    , VF_LINE (2,   1, -1, 0, 1)
    , VF_LINE (1,   1,  0, 0, 1)
    ,
    // Schwanz
    '\'', 2  RIGHT1
    , VF_LINE (1,   1,  1, 1, 1)
    , VF_LINE (1,   1,  0, 0, 1)
    ,
    // Tabellen-Ende
    VF_UNKNOWN_EXTRA, 1
    , VF_LINE (1,  1,  0, 1, 0)
};

