/*
    This file is part of SNAKE.

    SNAKE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    SNAKE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include <string.h>

#include "snake.h"

#define SNAKE_PIXEL_PER_FRAME 1200

#if defined  __AVR__
    // <stdio.h>
    #define printf(...) (void) 0
    #include <avr/io.h>
    #include <avr/pgmspace.h>
    #include <avr/interrupt.h>

    #include <pgmread-inc.h>
    #include <reload.h>

    #include "morse.h"
    #include "sound.h"
    #include "uput.h"
    #include "beam.h"

#elif defined __i386__

    // avr/io.h
    static uint8_t SREG; 
    // avr/pgmspace.h
    #define pgm_read_byte(x)  (*((const uint8_t*)  x))
    #define pgm_read_word(x)  (*((const uint16_t*) x))
    #define PROGMEM
    #define memcpy_P memcpy
    // avr/interrupt.h
    #define sei() (void) 0 
    #define cli() (void) 0 
    // pgmread-inc.h
    #define pgm_read_byte_inc(x)  (*x++)
    // morse.h
    #define morse(...) (void) 0
    // sound.h
    #define sound(...) (void) 0
    // uput.h
    #include <stdio.h>
    //#define printf(a,x...) fprintf (stderr, "#" a , ##x)
    #define printf(a,x...) fprintf (stderr, "%02x: " a, snake.draw.elem , ##x)
    // beam.h
    extern void beam_xy (vf_uint_t x, vf_uint_t y);
    extern void beam_skip (void);
    extern struct { uint16_t ticks; } beam;
    // SNAKE_PIXEL_PER_FRAME
    // Umdefinieren damit es aufm PC keine Totzeiten nach einem Frame gibt
    #undef  SNAKE_PIXEL_PER_FRAME
    #define SNAKE_PIXEL_PER_FRAME 1
    
    #include <stdlib.h>
    // reload.h
    #define RELOAD(...) (void) 0;
#else
#error
#endif // AVR, 386

#include "rc5-codes.h"
#include "parith-15.h"

snake_t snake;

static char snake_text[3][30];

static void snake_umlauf (snake_t*, uint8_t, uint8_t);
static void snake_check_bounds (snake_t * s, uint8_t dir, uint8_t len);
static void snake_frame_done (snake_t*);

// Richtungen
// NORT ^ SD  = 2
// WEST ^ OST  = 2
// ++ dreht nach links  (math. pos.)
// -- dreht nach rechts (math. neg.)
#define SN_N 3 // Nord
#define SN_S 1 // Sd
#define SN_W 0 // West
#define SN_E 2 // Ost

#define WALL_SKIP 2

#include "snake-data.h"

// Lnge der Schlang beim Start
#define START_LENGTH 20

// Tastendruck/Kommando
static void onKey_snake   (snake_t*, uint8_t code);

// Einen weiteren Pixel zeichnen
static void onPixel_snake (snake_t*);

// Menu-Punkt "Snake" wird aktiviert
static void onEnter_snake (snake_t*);

// tempus fugit
static void onTimeChanged_snake (snake_t*, long);

// Menu-Punkt "Snake"
const menuitem_t menuitem_snake PROGMEM = 
{
    .onPixel       = onPixel_snake,
    .onEnter       = onEnter_snake,
    .onTimeChanged = onTimeChanged_snake,
    .onKey         = onKey_snake,
    .use_fifo      = 1
};

// Tastendruck
void onKey_snake (snake_t * s, uint8_t code)
{
    if (s->game.over)
        return;
        
    uint8_t dir = s->dir;
    
    if (RC5_LEFT  == code)   dir = SN_W;
    if (RC5_RIGHT == code)   dir = SN_E;
    if (RC5_UP2   == code)   dir = SN_N;
    if (RC5_DOWN2 == code)   dir = SN_S;
    if (RC5_UP    == code)   dir = SN_N;
    if (RC5_DOWN  == code)   dir = SN_S;
    if (RC5_11 == code)      s->level.delay --;
    if (RC5_12 == code)      s->level.delay ++;
    
    dir &= 3;

    // Keine 180-Wende: Neue Richtung erst spter verarbeiten
    if (dir != s->body.seg[s->body.start].dir)
        s->new_dir = dir;
}

void onTimeChanged_snake (snake_t * s, long tid __attribute__((unused)))
{
    if (s->game.over)
    {
        if (s->draw.bums)
        {
            s->draw.bums = 0;
            s->scroll_y = 0;
            sound (1);
        }
        
        if (++ s->delay_scroll & 1)
            s->scroll_y++;
    }
    
    uint8_t delay = s->delay;
    if (--delay)
    {
        s->delay = delay;
        return;
    }
    s->delay = s->level.delay;
    
    s->changed.snake = 1;
}

// Schlange einen Pixel weiter
void move_snake (snake_t * s)
{
    uint8_t start = s->body.start;
    uint8_t nseg  = s->body.nseg;
    
    if (s->game.over)
    {
        if (!nseg)
            return;
        goto _schrumpf;
    }
        
    // Kopf 1 weiter
    uint8_t x = s->head.x;
    uint8_t y = s->head.y;
    
    uint8_t dir  = s->dir;
    uint8_t mask = s->raster_mask;
    
    // Falls Kopf auf Rasterpunkt: ggf. neue Richtung bernehmen
    if (s->new_dir != 0xff 
        && (x & mask) == 0 && (y & mask) == 0)
    {
        s->dir = dir = s->new_dir;
        s->new_dir = 0xff;
    }

    // Schritt in die Richtung
    if (dir & 1)    y += dir - 2;
    else            x += dir - 1;
    
    s->head.x = x;
    s->head.y = y;

    if (x == s->food.x  &&  y == s->food.y)
    {
        // Mampf!
        sound (0);
        s->grow = s->level.grow;
        s->changed.text = 1;
        s->game.score += s->level.grow;
        
        // neues Futter suchen gehen
        s->seek_food = 1;
        s->food_ok = 0;
    }
    
    // Erstes und letztes Segment anpassen
    
    uint8_t dirs = 2 ^ dir;
    
    if (dirs != s->body.seg[start].dir)
    {
        // Richtung wurde gewechselt -> ein neues Segment vorhngen
        nseg++;
        
        // start-- (mod SEG_MAX)
        
        if (--start == 0xff)
            start = SEG_MAX-1;

        // Start-Index hat sich gendert
        s->body.start = start;
        // Neues Segment initialisieren
        s->body.seg[start].len = 0;
        s->body.seg[start].dir = dirs;
    }
    
    // Erstes Segment 1 lnger
    s->body.seg[start].len += 1;

    if (nseg > SEG_MAX)
        // zu viele Segmente: abschneiden
        nseg = SEG_MAX;
    else
    {
        // Noch wachsen?
        if (s->grow)
            s->grow--;
        else
        {
        _schrumpf:;
            // Nicht mehr wachsen: letztes Segment 1 krzer
            // Ende 1 krzen
            // Index des letzten Segments (mod SEG_MAX)
            uint8_t idx  = start + nseg-1;
            if (idx >= SEG_MAX)
                idx -= SEG_MAX;
                
            uint8_t len = s->body.seg[idx].len - 1;
            
            if (len)
                // Segment 1 krzer
                s->body.seg[idx].len = len;
            else
                // Segment ist leer --> letztes Segment entfernen
                nseg--;
        }
    }

    if (s->body.nseg != nseg)
    {
        s->body.nseg = nseg;
        s->changed.text = 1;
    }
}

// Schwierigkeitsgrad einstellen
void snake_set_level (snake_t * s, uint8_t level)
{
    memcpy_P (& s->level, & SNAKE_LEVELS[level-1] , sizeof (snake_level_t));
    s->game.level = level;
    s->raster_mask = s->level.raster-1;
}

// Menu-Punkt "Snake" aktivieren
void onEnter_snake (snake_t * s)
{
    snake_set_level (s, 3);
    
    // Reset snake
    s->new_dir = 0xff;
    s->grow = 0;
    s->changed.snake = 1;
    s->changed.text  = 1;
    s->delay = 1;
    s->game.over = 0;
    s->game.score = 0;
    
    // Futter ist noch nicht gefunden
    s->seek_food = 1;
    s->food_ok = 0;
    
    // Schlong hat 1 Segment
    s->head.x = 0x80;
    s->head.y = 0x80;
    s->body.start = 0;
    s->body.nseg  = 1;
    s->body.seg[0].len = START_LENGTH;
    s->dir = prandom() & 3;
    s->body.seg[0].dir = 2 ^ s->dir;

    // Reset Pixel-Routine
    s->draw.done = 1;
    s->draw.bums = 0;

    // Text-Zeichner initialisieren    
    
    letter_drawinfo_t * di = & drawinfo;
    
    RELOAD ("r", di);

    di->beam_xy   = beam_xy;
    di->beam_skip = beam_skip;
    di->text_in_flash = 0;
    di->proportional = 1;

    draw_reset (di);
}

void snake_umlauf (snake_t * s, uint8_t dir, uint8_t len)
{
    if (dir & 1)
        // NORD || SD
        return;

    // Q: Testpunkt
    // P0/P1: Beginn/Ende des Mauerelements
    uint8_t yq = s->food.y;
    uint8_t y0 = s->draw.y;
    
    // Teststrahl soll nach oben gehen
    if (y0 < yq)
        return;
        
    // Trifft der Strahl die Mauer?    
    uint8_t xq = s->food.x;
    uint8_t x0 = s->draw.x;
    uint8_t x1 = x0;
    
    if (dir == SN_E)
    {
        // Ost
        x1 += len;
        if (xq < x0 || xq > x1)
            // Nein
            return;
    }
    else
    {
        // West
        x1 -= len;
        if (xq < x1 || xq > x0)
            // Nein
            return;
    }
    
    // Ja, der Strahl trifft
    // WEST: +1
    // Ost:  -1
    s->wall_umlauf -= dir -1;
}

void snake_check_bounds (snake_t * s, uint8_t dir, uint8_t len)
{
    if (s->game.over)
        return;
        
    if (s->seek_food)
        snake_umlauf (s, dir, len);
    // In .draw wurden soeben .x, .y, .dir und .len fr ein
    // Mauerstck gesetzt
    
    // Beginn des Mauerelements
    uint8_t x = s->draw.x;
    uint8_t y = s->draw.y;
    
    // ka darf nicht 0 sein falls kb im Intervall [k0,k1]
    uint8_t ka, kb;
    uint8_t k0, k1;

    if (dir & 1)
    {
        // Mauer in y-Richtung 
        ka = s->head.x - x;
        kb = s->head.y;
        k0 = y;
    }
    else
    {
        // Mauer in x-Richtung 
        ka = s->head.y - y;
        kb = s->head.x;
        k0 = x;
    }
    
    k1 = k0;

    // Punkt neben dem Mauerstck?
    if (dir >= 2)
    {
        // NORD || OST
        k1 += len;
        if (kb < k0 || kb > k1)     return;     // Nein
    }
    else
    {
        // WEST || SD
        k1 -= len;
        if (kb < k1 || kb > k0)     return;     // Nein
    }
    // Ja
    // Genau drauf?
    if (ka)    return;  // Nein
    // Ja
    // Bums!
    s->draw.bums = 1;
    s->game.over = 1;
}

void snake_frame_done (snake_t * s)
{
    // Alles ist fertig gezeichnet:
    uint8_t sreg = SREG;
    cli();
    uint16_t ticks = beam.ticks;
    SREG = sreg;

    // Nichts tun bis ticks = SNAKE_PIXEL_PER_FRAME
    // Damit bei wenigen zu zeichnenden Pixeln diese
    // nicht heller sind.
    if (ticks < SNAKE_PIXEL_PER_FRAME)
    {
        // 1000 Ticks bei 37.5 kPix/s --> 37.5 fps
        if (s->draw.done == 1)
        {
            // Strahl austasten und ins Aus stellen
            beam_skip();
            beam_xy (255, 255);
            s->draw.done = 2;
        }
        return;
    }

    // Ticks wieder auf 0
    cli();
    beam.ticks = 0;
    SREG = sreg;

    // Zeichnen beginnt von vorne
    s->draw.done = 0;
    s->draw.len  = 0;
    s->draw.elem = 0xfa-1;
    s->draw.ntext = 0;
    s->ntexte = 2 + s->game.over;

    // onTimeChanged wurde aufgerufen: Schlange 1 weiter oder schrumpfen
    if (s->changed.snake)
    {
        s->changed.snake = 0;
        move_snake (s);
    }
    // Text hat sich gendert: Strings neu schreiben
    if (s->changed.text)
    {
        s->changed.text = 0;
        
#if defined __AVR__
        put_P ("%@Score: %u\nLevel: %u"
               "%@Game over!",
               snake_text[0], s->game.score, s->game.level,
               snake_text[2]);
#elif defined __i386__
        sprintf (snake_text[0], "Score: %u\nLevel: %u",
                 s->game.score, s->game.level);
                            
        sprintf (snake_text[2], "Game over!");
#else
#error
#endif // AVR, 386, ...

    }
    // Falls Futter nicht platziert ist:
    // Einen neuen, zuflligen Ort erzeugen
    if (s->seek_food)
    {
        if (s->food_ok)
            s->seek_food = 0;
        else
        {
            uint8_t mask = 0x7f & ~s->raster_mask;
            s->food.x = 0x40 + (uint8_t) (prandom() & mask);
            s->food.y = 0x40 + (uint8_t) (prandom() & mask);
        }
    }
}

// return skip
static uint8_t read_wall_seg (snake_t *s)
{
    const uint8_t * wall = s->draw.wall;
    uint8_t dir_len, dir, len;
    
    RELOAD ("y", s);
 _next_seg:;    
    uint8_t first = 0;
    dir_len  = pgm_read_byte_inc (wall);

    if (dir_len == WALL_GOTO)
    {
        first = 1;
        s->draw.x = pgm_read_byte_inc (wall);
        s->draw.y = pgm_read_byte_inc (wall);
        dir_len   = pgm_read_byte_inc (wall);
    }
    
    // dir.1 steht in Bit 0. Da WALL auf gerade Adressen aligned ist und
    // das erste Element nach Osten geht, ergibt sich dir.0 aus Offset.0,
    // was wiederum wall.0 ist.
    dir = dir_len & 1;
    dir <<= 1;
    // Abfrage umdrehen wegen dem *wall++ von eben
    uint8_t bit0;
    
#define AAA(x) ((unsigned int) x)
    if (bit0 = AAA(wall) ^ AAA(WALL), bit0 % 2 == 1)
        dir |= 1;
        
    uint8_t level = dir_len >> 1;
    level &= 0x43; // 0b01000011
    
    uint8_t level_ok = s->game.level > level 
        || (uint8_t) (s->game.level + 0x40) <= level;

    // level entfernen    
    dir_len &= ~0x86; // 0b01111001
    s->draw.elem = 0xfd-1;

    if (dir_len == 1)
    {
        // noch eine Mauer
        // Mauer in diesem Level aktiv?
        if (level_ok)
        {
            // Ja.
            s->draw.wall = wall;
            s->draw.len  = 0;
            return 1;
        }
        // Nein --> len = 0 beendet Mauer-Ausgabe
        dir_len = 0;
    }
    
    len = dir_len & 0x78; // b01111000

    // Segment ab diesem Level aktiv?
    if (len && !level_ok)
        // Nein
        goto _next_seg;
    
    s->draw.wall = wall;
    s->draw.dir  = dir;

    if (!len)
    {
        // Alle Mauern sind fertig
        s->draw.len = 0;
        
        if (s->seek_food)
            // Berechnung der Umlaufzahl fr Futter ist damit abgeschlossen
            s->food_ok = s->wall_umlauf == 1;

        s->draw.elem = 0xfd;
        return 1;
    }

    // Noch ein Durchlauf...
    s->draw.elem = 0xfd-1;
    
    snake_check_bounds (s, dir, len);
        
    if (first)
    {
        // Das erste Element (es geht nach Osten) 1 lnger machen
        // um den Strahl zu schleppen.
        // Verlnderung erst nach snake_check_bounds
        s->draw.len  = len + (1 << WALL_SKIP);
        s->draw.x   -= (1 << WALL_SKIP);
        return 1;
    }

    s->draw.len = len;
    return 0;
}

// 1 Pixel aufs Rohr
void onPixel_snake (snake_t * s)
{
    if (s->draw.done)
        return snake_frame_done (s);
    
    // Texte, Mauer, Schlange (Kopf, Body), Futter, etc. zeichnen
    // elem = 
    //    TEXT <-- fa, fb
    //    WALL <-- fc, fd
    //    FOOD <-- fe
    //    HEAD <-- ff
    //    BODY <-- 0...nseg-1

    // # noch zu zeichnender Pixel fr dieses Element
    uint8_t len  = s->draw.len;
    uint8_t x    = s->draw.x;
    uint8_t y    = s->draw.y;
    uint8_t elem = s->draw.elem;

    if (len == 0)
    {
    _next_element:;
        // Elemenet fertig: nchstes Segment
        
        if (++elem == s->body.nseg)
        {
            // Letztes Element ist gezeichnet: dieser Frame ist fertig
            s->draw.done = 1;
            return beam_skip();
        }
        
        if (elem >= 0xfe)
        {
            // Koordinaten fr Kopf, Futter, ...
            uint8_t idx = ~elem;
            s->draw.x = s->p[idx].x;
            s->draw.y = s->p[idx].y;
        }
        
        s->draw.elem = elem;
        
        // Springt der Strahl?
        uint8_t skip = 1;

        if (elem < 0x80)
        {
            //  Krper-Segment
            uint8_t idx = elem + s->body.start;
            
            if (idx >= SEG_MAX)
                idx -= SEG_MAX;
            
            // Richtung und Lnge
            s->draw.dir = s->body.seg[idx].dir;
            len = s->body.seg[idx].len;
            skip = 0;
        }
        else if (elem == 0xff-SN_HEAD)
        {
            if (!s->body.nseg)
                goto _next_element;
            // Kopf
            s->draw.dir = s->dir;
            len = 7;
            skip = 1;
        }
        else if (elem == 0xff-SN_FOOD)
        {
            // Futter noch nicht platziert
            if (s->seek_food || s->game.over)
                goto _next_element;
            len = 9;
            skip = 1;
        }
        else if (elem == 0xfa)
        {
            // Text init
            uint8_t n = s->draw.ntext;
            const uint8_t * tinit = (const uint8_t*) & snake_text_init[n];
            drawinfo.start.x = pgm_read_byte_inc (tinit);
            drawinfo.start.y = (n == 2)
                ? tinit++, s->scroll_y
                : pgm_read_byte_inc (tinit);
            drawinfo.zoom    = pgm_read_byte_inc (tinit);
            drawinfo.text    = snake_text[n];
            if (n == 1)
            {
                char * p = snake_text[1];
                RELOAD ("e", p);
                *p++ = '$';
                for (uint8_t i = s->body.nseg; i; i--)
                    *p++ = '~';
                *p++ = '\'';
                *p = '\0';
            }
                
            return beam_skip();
        }
        else if (elem == 0xfb)
        {
            // Text
            s->draw.elem = --elem;
            skip = 0;
        }
        else if (elem == 0xfc)
        {
            // Mauer-Anfang
            s->wall_umlauf = 0;
            s->draw.wall = WALL;
            read_wall_seg (s);
            //s->draw.elem = 0xfd-1;
            return beam_skip();
        }
        else if (elem == 0xfd)
        {
            // Mauer
            if (read_wall_seg (s))
                return beam_skip();

            // read_wall_seg verndert x, y, len: aktualisieren
            x    = s->draw.x;
            y    = s->draw.y;
            len  = s->draw.len;
            elem--;
            skip = 0;
        }
        
        if (skip)
        {
            s->draw.len = len;
            // Strahl austasten
            return beam_skip();
        }
    }

    if (elem == 0xfb-1)
    {
        // Text
        uint8_t text_done = draw_letters (& drawinfo);
        s->draw.len = !text_done;
        if (text_done)
            s->draw.elem = (++ s->draw.ntext >= s->ntexte)
                ? ++elem
                : --elem;
        return;
    }
    
    uint8_t dir  = s->draw.dir;

    if (elem == 0xfd-1)
    {
        // Mauer zeichnen
        // (1 << WALL_SKIP) Schritte in Segment-Richtung fortschreiten
        s->draw.len = len - (1 << WALL_SKIP);
        uint8_t ndir = dir << WALL_SKIP;
        if (dir & 1)    y += ndir - (2 << WALL_SKIP);
        else            x += ndir - (1 << WALL_SKIP);
            
        s->draw.x = x;
        s->draw.y = y;

        // Wiederhole; read_wall_seg() kmmert sich um Mauer-Ende
        s->draw.elem = 0xfd-1;

        // Punkt auf die Rhre    
        return beam_xy (x, y);
    }
    
    s->draw.len = --len;
    
    if (elem >= 0xfe)
    {
        // Pixel von Kopf, Futter, ... zeichnen
        uint16_t dxy = pgm_read_word (& snake_head_dxy[len]);
        
        uint8_t dx = dxy;
        uint8_t dy = dxy >> 8;
        
        if (dir < 2)            dx = -dx, dy = -dy;
        if (!(dir & 1))         dx ^= dy, dy ^= dx, dx ^= dy;
        
        x += dx;
        y += dy;
    }
    else if (elem < 0x80)
    {
        // Body zeichnen
        // 1 Schritt in Segment-Richtung fortschreiten
        if (dir & 1)    y += dir-2;
        else            x += dir-1;
            
        s->draw.x = x;
        s->draw.y = y;
        
        if (s->seek_food && s->food_ok)
        {
            // Neues Futter wurde erzeugt
            if (!s->level.food_on_snake)
                // Futter ist auf Body        
                if (x == s->food.x && y == s->food.y)
                    s->food_ok = 0;
        }

        // Kopf beit in den Body        
        if (!s->game.over && x == s->head.x && y == s->head.y)
        {
            s->draw.bums = 1;
            s->game.over = 1;
        }
    }

    // Punkt auf die Rhre    
    beam_xy (x, y);
}
